"""Define the SmartThings Cloud API."""

from typing import List, Optional, Sequence

from .api import API
from .app import App, AppEntity, AppSettings, AppSettingsEntity
from .device import DeviceEntity
from .installedapp import InstalledAppEntity
from .location import LocationEntity
from .oauth import OAuth, OAuthClient, OAuthEntity
from .subscription import Subscription, SubscriptionEntity


class SmartThings:
    """Define a class for interacting with the SmartThings Cloud API."""

    def __init__(self, token):
        """
        Initialize the API.

        :param token: The personal access token used to authenticate to the API
        :type token: str
        """
        self._api = API(token)

    def locations(self) -> List:
        """Retrieve SmartThings locations."""
        resp = self._api.get_locations()
        return [LocationEntity(self._api, entity) for entity in resp["items"]]

    def location(self, location_id: str) -> LocationEntity:
        """Retrieve a location with the specified ID."""
        entity = self._api.get_location(location_id)
        return LocationEntity(self._api, entity)

    def devices(self, location_ids: Optional[Sequence[str]] = None,
                capabilities: Optional[Sequence[str]] = None,
                device_ids: Optional[Sequence[str]] = None) -> List:
        """Retrieve SmartThings devices."""
        params = {}
        if location_ids:
            params['locationId'] = location_ids
        if capabilities:
            params['capability'] = capabilities
        if device_ids:
            params['deviceId'] = device_ids
        resp = self._api.get_devices(params)
        return [DeviceEntity(self._api, entity) for entity in resp["items"]]

    def device(self, device_id: str) -> DeviceEntity:
        """Retrieve a device with the specified ID."""
        entity = self._api.get_device(device_id)
        return DeviceEntity(self._api, entity)

    def apps(self) -> List[AppEntity]:
        """Retrieve list of apps."""
        resp = self._api.get_apps()
        return [AppEntity(self._api, entity) for entity in resp["items"]]

    def app(self, app_id: str) -> AppEntity:
        """Retrieve an app with the specified ID."""
        entity = self._api.get_app(app_id)
        return AppEntity(self._api, entity)

    def create_app(self, app: App) -> (AppEntity, OAuthClient):
        """Create a new app."""
        entity = self._api.create_app(app.to_data())
        return AppEntity(self._api, entity['app']), OAuthClient(entity)

    def delete_app(self, app_id: str):
        """Delete an app."""
        return self._api.delete_app(app_id) == {}

    def app_settings(self, app_id: str) -> AppSettingsEntity:
        """Get an app's settings."""
        return AppSettingsEntity(
            self._api, app_id, self._api.get_app_settings(app_id))

    def update_app_settings(self, data: AppSettings) -> AppSettingsEntity:
        """Update an app's settings."""
        entity = self._api.update_app_settings(data.app_id, data.to_data())
        return AppSettingsEntity(self._api, data.app_id, entity)

    def app_oauth(self, app_id: str) -> OAuthEntity:
        """Get an app's OAuth settings."""
        return OAuthEntity(self._api, app_id, self._api.get_app_oauth(app_id))

    def update_app_oauth(self, data: OAuth) -> OAuthEntity:
        """Update an app's OAuth settings without having to retrieve it."""
        entity = self._api.update_app_oauth(data.app_id, data.to_data())
        return OAuthEntity(self._api, data.app_id, entity)

    def installedapps(self) -> List[InstalledAppEntity]:
        """Get a list of the installed applications."""
        resp = self._api.get_installedapps()
        return [InstalledAppEntity(self._api, entity)
                for entity in resp["items"]]

    def installedapp(self, installed_app_id: str) -> InstalledAppEntity:
        """Get an installedapp with the specified ID."""
        entity = self._api.get_installedapp(installed_app_id)
        return InstalledAppEntity(self._api, entity)

    def delete_installedapp(self, installed_app_id: str):
        """Delete an installedapp."""
        return self._api.delete_installedapp(installed_app_id) == {'count': 1}

    def subscriptions(self, installed_app_id: str) \
            -> List[SubscriptionEntity]:
        """Get an installedapp's subscriptions."""
        resp = self._api.get_subscriptions(installed_app_id)
        return [SubscriptionEntity(self._api, entity)
                for entity in resp["items"]]

    def delete_subscriptions(self, installed_app_id: str) -> int:
        """Delete an installedapp's subscriptions."""
        resp = self._api.delete_all_subscriptions(installed_app_id)
        return resp['count']

    def delete_subscription(self, installed_app_id: str,
                            subscription_id: str):
        """Delete an individual subscription."""
        return self._api.delete_subscription(
            installed_app_id, subscription_id) == {'count': 1}

    def create_subscription(self, subscription: Subscription) \
            -> SubscriptionEntity:
        """Create a new subscription for an installedapp."""
        entity = self._api.create_subscription(
            subscription.installed_app_id, subscription.to_data())
        return SubscriptionEntity(self._api, entity)



"""A Mock API for SmartThings."""

from collections import namedtuple

from requests import Request, Response
from requests_mock.response import create_response

from pysmartthings import api, oauthapi

from .utilities import get_json

API_TOKEN = "Test Token"
APP_ID = 'c6cde2b0-203e-44cf-a510-3b3ed4706996'
DEVICE_ID = '743de49f-036f-4e9c-839a-2f89d57607db'
INSTALLED_APP_ID = '4514eb36-f5fd-4ab2-9520-0597acd1d212'
SUBSCRIPTION_ID = '7bdf5909-57c4-41f3-9089-e520513bd92a'
LOCATION_ID = '397678e5-9995-4a39-9d9f-ae6ba310236b'
CLIENT_ID = '7cd4d474-7b36-4e03-bbdb-4cd4ae45a2be'
CLIENT_SECRET = '9b3fd445-42d6-441b-b386-99ea51e13cb0'
REFRESH_TOKEN = 'a86a5c8e-0014-44a6-8980-5846633972dd'

UrlMock = namedtuple('UrlMock', 'method url request response')

URLS = [
    UrlMock('GET', api.API_LOCATIONS, None, 'locations.json'),
    UrlMock('GET', api.API_LOCATION.format(location_id=LOCATION_ID),
            None, 'location.json'),
    UrlMock('GET', api.API_DEVICES, None, 'devices.json'),
    UrlMock('GET', api.API_DEVICES +
            "?locationId=397678e5-9995-4a39-9d9f-ae6ba310236b" +
            "&capability=switch" +
            "&deviceId=edd26ac6-d156-4505-9647-3b20118ae4d1" +
            "&deviceId=be1a61ce-c2a4-4b32-bf8c-31de6d3fa7dd",
            None, 'devices_filtered.json'),
    UrlMock('GET', api.API_DEVICE.format(device_id=DEVICE_ID),
            None, 'device.json'),
    UrlMock('GET', api.API_DEVICE_STATUS.format(device_id=DEVICE_ID),
            None, 'device_status.json'),
    UrlMock('POST', api.API_DEVICE_COMMAND.format(device_id=DEVICE_ID),
            'device_command_post_switch_on.json', {}),
    UrlMock('POST', api.API_DEVICE_COMMAND.format(device_id=DEVICE_ID),
            'device_command_post_switch_off.json', {}),
    UrlMock('POST', api.API_DEVICE_COMMAND.format(device_id=DEVICE_ID),
            'device_command_post_set_level.json', {}),
    UrlMock('GET', api.API_APPS, None, 'apps.json'),
    UrlMock('GET', api.API_APP.format(app_id=APP_ID),
            None, 'app_get.json'),
    UrlMock('POST', api.API_APPS,
            'app_post_request.json', 'app_post_response.json'),
    UrlMock('PUT', api.API_APP.format(app_id=APP_ID),
            'app_put_request.json', 'app_put_response.json'),
    UrlMock('DELETE', api.API_APP.format(app_id=APP_ID), None, {}),
    UrlMock('GET', api.API_APP_SETTINGS.format(app_id=APP_ID),
            None, 'app_settings.json'),
    UrlMock('PUT', api.API_APP_SETTINGS.format(app_id=APP_ID),
            'app_settings.json', 'app_settings.json'),
    UrlMock('GET', api.API_APP_OAUTH.format(app_id=APP_ID),
            None, 'app_oauth_get_response.json'),
    UrlMock('PUT', api.API_APP_OAUTH.format(app_id=APP_ID),
            'app_oauth_put_request.json', 'app_oauth_put_response.json'),
    # InstalledApps
    UrlMock('GET', api.API_INSTALLEDAPPS, None,
            'installedapps_get_response.json'),
    UrlMock('GET', "https://api.smartthings.com/installedapps?"
                   "currentLocationId=NWMwM2U1MTgtMTE4YS00NGNiLTg1YWQtNzg3N2Qw"
                   "YjMwMmU0&currentOffset=MA", None,
            'installedapps_get_response_2.json'),
    UrlMock('GET', api.API_INSTALLEDAPP.format(
        installed_app_id=INSTALLED_APP_ID),
            None, 'installedapp_get_response.json'),
    UrlMock('DELETE', api.API_INSTALLEDAPP.format(
        installed_app_id=INSTALLED_APP_ID), None, {"count": 1}),
    UrlMock('GET', api.API_SUBSCRIPTIONS.format(
        installed_app_id=INSTALLED_APP_ID), None,
            'subscriptions_get_response.json'),
    UrlMock('DELETE', api.API_SUBSCRIPTIONS.format(
        installed_app_id=INSTALLED_APP_ID), None,
            {'count': 3}),
    UrlMock('POST', api.API_SUBSCRIPTIONS.format(
        installed_app_id=INSTALLED_APP_ID),
            'subscription_post_request.json',
            'subscription_post_response.json'),
    UrlMock('GET', api.API_SUBSCRIPTION.format(
        installed_app_id=INSTALLED_APP_ID,
        subscription_id='7bdf5909-57c4-41f3-9089-e520513bd92a'), None,
            'subscription_capability_get_response.json'),
    UrlMock('GET', api.API_SUBSCRIPTION.format(
        installed_app_id=INSTALLED_APP_ID,
        subscription_id='498752fd-db87-4a5e-95f5-25a0e412838d'), None,
            'subscription_device_get_response.json'),
    UrlMock('DELETE', api.API_SUBSCRIPTION.format(
        installed_app_id=INSTALLED_APP_ID,
        subscription_id=SUBSCRIPTION_ID), None,
            {'count': 1})
]


def setup(requests_mock):
    """Configure request mocks the API calls."""
    requests_mock.add_matcher(__matcher)

    requests_mock.register_uri(
        'POST', oauthapi.AUTH_API_TOKEN,
        json=get_json('token_response.json'))


def __matcher(req: Request) -> Response:
    """Match against our registry."""
    match = next((obj for obj in URLS if __match_request(req, obj)), None)
    if match:
        return create_response(req, json=__get_body(match.response))


def __match_request(req: Request, mock: UrlMock):
    """Match the request against the mock setup."""
    if not req.headers.get('Authorization', '') == "Bearer " + API_TOKEN:
        return False
    if not req.method == mock.method:
        return False
    target_url = mock.url if mock.url.startswith('http') \
        else api.API_BASE + mock.url
    if not req.url == target_url:
        return False
    if mock.request and not req.json() == __get_body(mock.request):
        return False
    return True


def __get_body(body):
    if isinstance(body, (dict, list)):
        return body
    if isinstance(body, str):
        return get_json(body)
    return None



"""Defines the entity module."""

from .api import API


class Entity:
    """Define an entity from the SmartThings API."""

    _api: API

    def __init__(self, api: API):
        """Initialize a new instance of the entity."""
        self._api = api

    def refresh(self):
        """Retrieve the latest values from the API."""
        raise NotImplementedError

    def save(self):
        """Update or create the entity."""
        raise NotImplementedError



"""Define the OAuthAPI module."""

import requests

from .errors import APIInvalidGrant, APIUnauthorizedError, APIUnknownError

AUTH_API_TOKEN = "https://auth-global.api.smartthings.com/oauth/token"


class OAuthAPI:
    """Define the OAuthAPI class."""

    def __init__(self, client_id: str, client_secret: str):
        """Initialize a new instance of the OAuthAPI."""
        self._client_id = client_id
        self._client_secret = client_secret

    def get_token(self, refresh_token: str):
        """Obtain a new access and refresh token."""
        payload = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }
        resp = requests.post(AUTH_API_TOKEN,
                             auth=(self._client_id, self._client_secret),
                             data=payload)
        if resp.ok:
            return resp.json()
        if resp.status_code == 400:
            data = resp.json()
            raise APIInvalidGrant(data['error_description'])
        if resp.status_code == 401:
            raise APIUnauthorizedError
        raise APIUnknownError



"""Utility for invoking the SmartThings Cloud API."""

from typing import Optional

import requests

from . import errors

API_BASE: str = "https://api.smartthings.com/v1/"
API_LOCATIONS = "locations"
API_LOCATION = API_LOCATIONS + "/{location_id}"
API_DEVICES: str = "devices"
API_DEVICE: str = API_DEVICES + "/{device_id}"
API_DEVICE_STATUS: str = "devices/{device_id}/status"
API_DEVICE_COMMAND: str = "devices/{device_id}/commands"
API_APPS: str = "apps"
API_APP: str = "apps/{app_id}"
API_APP_OAUTH: str = "apps/{app_id}/oauth"
API_APP_SETTINGS: str = "apps/{app_id}/settings"
API_INSTALLEDAPPS: str = "installedapps"
API_INSTALLEDAPP: str = "installedapps/{installed_app_id}"
API_SUBSCRIPTIONS: str = API_INSTALLEDAPP + "/subscriptions"
API_SUBSCRIPTION: str = API_SUBSCRIPTIONS + "/{subscription_id}"


class API:
    """
    Utility for invoking the SmartThings Cloud API.

    https://smartthings.developer.samsung.com/docs/api-ref/st-api.html
    """

    def __init__(self, token: str):
        """Initialize a new instance of the API class."""
        self._headers = {"Authorization": "Bearer " + token}

    def get_locations(self) -> dict:
        """
        Get locations.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/listLocations
        """
        return self._request_paged_list('get', API_LOCATIONS)

    def get_location(self, location_id: str) -> dict:
        """
        Get a specific location.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getLocation
        """
        return self._request(
            'get', API_LOCATION.format(location_id=location_id))

    def get_devices(self, params: Optional[dict] = None) -> dict:
        """
        Get the device definitions.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getDevices
        """
        return self._request_paged_list('get', API_DEVICES, params=params)

    def get_device(self, device_id: str) -> dict:
        """
        Get as specific device.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getDevice
        """
        return self._request(
            'get', API_DEVICE.format(device_id=device_id))

    def get_device_status(self, device_id: str) -> dict:
        """Get the status of a specific device."""
        return self._request(
            'get',
            API_DEVICE_STATUS.format(device_id=device_id))

    def post_command(self, device_id, capability, command, args,
                     component="main") -> object:
        """
        Execute commands on a device.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/executeDeviceCommands
        """
        data = {
            "commands": [
                {
                    "component": component,
                    "capability": capability,
                    "command": command
                }
            ]
        }
        if args:
            data["commands"][0]["arguments"] = args

        return self._request(
            'post',
            API_DEVICE_COMMAND.format(device_id=device_id),
            data)

    def get_apps(self) -> dict:
        """
        Get list of apps.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/listApps
        """
        return self._request_paged_list('get', API_APPS)

    def get_app(self, app_id: str) -> dict:
        """
        Get the details of the specific app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/getApp
        """
        return self._request(
            'get',
            API_APP.format(app_id=app_id))

    def create_app(self, data: dict) -> dict:
        """
        Create a new app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/createApp
        """
        return self._request('post', API_APPS, data)

    def update_app(self, app_id: str, data: dict) -> dict:
        """
        Update an existing app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/updateApp
        """
        return self._request(
            'put', API_APP.format(app_id=app_id), data)

    def delete_app(self, app_id: str):
        """
        Delete an app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/deleteApp
        """
        return self._request(
            'delete', API_APP.format(app_id=app_id))

    def get_app_settings(self, app_id: str) -> dict:
        """
        Get an app's settings.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getAppSettings
        """
        return self._request('get', API_APP_SETTINGS.format(app_id=app_id))

    def update_app_settings(self, app_id: str, data: dict) -> dict:
        """
        Update an app's settings.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/updateAppSettings
        """
        return self._request(
            'put', API_APP_SETTINGS.format(app_id=app_id), data)

    def get_app_oauth(self, app_id: str) -> dict:
        """
        Get an app's oauth settings.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/getAppOauth
        """
        return self._request('get', API_APP_OAUTH.format(app_id=app_id))

    def update_app_oauth(self, app_id: str, data: dict) -> dict:
        """
        Update an app's oauth settings.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/updateAppOauth
        """
        return self._request(
            'put', API_APP_OAUTH.format(app_id=app_id), data)

    def get_installedapps(self) -> dict:
        """
        Get list of installedapps.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/listInstallations
        """
        return self._request_paged_list('get', API_INSTALLEDAPPS)

    def get_installedapp(self, installed_app_id: str) -> dict:
        """
        Get the details of the specific installedapp.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getInstallation
        """
        return self._request(
            'get',
            API_INSTALLEDAPP.format(installed_app_id=installed_app_id))

    def delete_installedapp(self, installed_app_id: str):
        """
        Delete an app.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/deleteInstallation
        """
        return self._request(
            'delete', API_INSTALLEDAPP.format(
                installed_app_id=installed_app_id))

    def get_subscriptions(self, installed_app_id: str) -> dict:
        """
        Get installedapp's subscriptions.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/listSubscriptions
        """
        return self._request_paged_list(
            'get',
            API_SUBSCRIPTIONS.format(installed_app_id=installed_app_id))

    def create_subscription(self, installed_app_id: str, data: dict) -> dict:
        """
        Create a subscription for an installedapp.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/saveSubscription
        """
        return self._request(
            'post',
            API_SUBSCRIPTIONS.format(installed_app_id=installed_app_id),
            data)

    def delete_all_subscriptions(self, installed_app_id: str) -> dict:
        """
        Delete all subscriptions for an installedapp.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/deleteAllSubscriptions
        """
        return self._request(
            'delete',
            API_SUBSCRIPTIONS.format(installed_app_id=installed_app_id))

    def get_subscription(self, installed_app_id: str, subscription_id: str) \
            -> dict:
        """
        Get an individual subscription.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/getSubscription
        """
        return self._request(
            'get',
            API_SUBSCRIPTION.format(
                installed_app_id=installed_app_id,
                subscription_id=subscription_id))

    def delete_subscription(self, installed_app_id: str, subscription_id: str):
        """
        Delete an individual subscription.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/deleteSubscription
        """
        return self._request(
            'delete',
            API_SUBSCRIPTION.format(
                installed_app_id=installed_app_id,
                subscription_id=subscription_id))

    @staticmethod
    def _get_next_link(data):
        links = data.get('_links')
        if not links:
            return None
        next_link = links.get('next')
        if not next_link:
            return None
        return next_link.get('href')

    def _request_paged_list(self, method: str, resource: str,
                            data: dict = None, params: dict = None):
        response = self._request(method, resource, data, params)
        items = response['items']
        next_link = API._get_next_link(response)
        while next_link:
            response = self._request(method, data=data, params=params,
                                     url=next_link)
            items.extend(response['items'])
            next_link = API._get_next_link(response)
        return {'items': items}

    def _request(self, method: str, resource: str = None, data: dict = None,
                 params: dict = None, *, url: str = None):
        response = requests.request(
            method,
            url if url else API_BASE + resource,
            params=params,
            json=data,
            headers=self._headers)

        if response.ok:
            return response.json()
        if response.status_code == 401:
            raise errors.APIUnauthorizedError
        elif response.status_code == 403:
            raise errors.APIForbiddenError
        raise errors.APIUnknownError



"""Tests for the OAuth module."""

import pytest

from pysmartthings.api import API
from pysmartthings.oauth import OAuth, OAuthEntity, OAuthToken
from pysmartthings.oauthapi import OAuthAPI

from . import api_mock
from .utilities import get_json


class TestOAuth:
    """Tests for the OAuth class."""

    @staticmethod
    def test_init():
        """Tests the initialization."""
        # Arrange
        app_id = '5c03e518-118a-44cb-85ad-7877d0b302e4'
        # Act
        oauth = OAuth(app_id)
        # Assert
        assert app_id == oauth.app_id
        assert oauth.scope is not None

    @staticmethod
    def test_client_name():
        """Tests get/set of client name."""
        # Arrange
        oauth = OAuth('5c03e518-118a-44cb-85ad-7877d0b302e4')
        # Act
        expected = "my_app"
        oauth.client_name = expected
        actual = oauth.client_name
        # Assert
        assert actual == expected

    @staticmethod
    def test_client_name_invalid():
        """Tests setting an invalid client name."""
        # Arrange
        oauth = OAuth('5c03e518-118a-44cb-85ad-7877d0b302e4')
        # Act/Assert
        with pytest.raises(ValueError):
            oauth.client_name = ''


class TestOAuthEntity:
    """Tests for the OAuthEntity class."""

    @staticmethod
    def test_refresh(requests_mock):
        """Tests the refresh method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        entity = OAuthEntity(
            api, 'c6cde2b0-203e-44cf-a510-3b3ed4706996', None)
        # Act
        entity.refresh()
        # Assert
        assert entity.client_name == 'pysmartthings-test'
        assert 'r:devices' in entity.scope

    @staticmethod
    def test_save(requests_mock):
        """Tests the refresh method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        entity = OAuthEntity(
            api, 'c6cde2b0-203e-44cf-a510-3b3ed4706996', None)
        entity.client_name = 'pysmartthings-test'
        entity.scope.append('r:devices')
        # Act/Assert
        entity.save()


class TestOAuthToken:
    """Tests for the OAuthToken class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange/Act
        token = OAuthToken(None, None)
        # Assert
        assert token.expires_in == 0
        assert token.scope == []

    @staticmethod
    def test_apply_data():
        """Tests the apply data method."""
        # Arrange
        data = get_json('token_response.json')
        # Act
        token = OAuthToken(None, data)
        # Assert
        assert token.expires_in == 299
        assert token.refresh_token == '3d1a8d0a-a312-45c2-a9f5-95e59dc0e879'
        assert token.access_token == 'ad0fbf27-48d4-4ee9-ba47-7f5fedd7be35'
        assert token.token_type == 'bearer'
        assert token.scope == ['r:devices:*']

    @staticmethod
    def test_refresh(requests_mock):
        """Tests the refresh method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = OAuthAPI(api_mock.CLIENT_ID, api_mock.CLIENT_SECRET)
        token = OAuthToken(api, refresh_token=api_mock.REFRESH_TOKEN)
        # Act
        token.refresh()
        # Assert
        assert token.refresh_token == '3d1a8d0a-a312-45c2-a9f5-95e59dc0e879'



"""Tests for the SmartThings file."""

from pysmartthings import create
from pysmartthings.app import App, AppSettings
from pysmartthings.oauth import OAuth
from pysmartthings.smartthings import SmartThings
from pysmartthings.subscription import Subscription

from . import api_mock
from .utilities import get_json


class TestSmartThings:
    """Tests for the SmartThings class."""

    @staticmethod
    def test_create(requests_mock):
        """Tests the create method."""
        # Arrange
        api_mock.setup(requests_mock)
        # Act
        smartthings = create(api_mock.API_TOKEN)
        devices = smartthings.devices()
        # assert
        assert len(devices) == 4

    @staticmethod
    def test_devices(requests_mock):
        """Tests devices are retrieved."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        devices = smartthings.devices()
        # Assert
        assert len(devices) == 4

    @staticmethod
    def test_devices_with_filter(requests_mock):
        """Tests retrieving a filtered view of devices."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        devices = smartthings.devices(
            location_ids=api_mock.LOCATION_ID,
            capabilities='switch',
            device_ids=['edd26ac6-d156-4505-9647-3b20118ae4d1',
                        'be1a61ce-c2a4-4b32-bf8c-31de6d3fa7dd'])
        # Assert
        assert len(devices) == 2

    @staticmethod
    def test_device(requests_mock):
        """Tests the device(id) method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        device = smartthings.device(api_mock.DEVICE_ID)
        # Assert
        assert device.device_id == api_mock.DEVICE_ID

    @staticmethod
    def test_locations(requests_mock):
        """Tests locations are retrieved."""
        # arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # act
        locations = smartthings.locations()
        # assert
        assert len(locations) == 2

    @staticmethod
    def test_location(requests_mock):
        """Tests the location(id) method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        location = smartthings.location(api_mock.LOCATION_ID)
        # Assert
        assert location.location_id == api_mock.LOCATION_ID

    @staticmethod
    def test_apps(requests_mock):
        """Tests locations are retrieved."""
        # arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # act
        apps = smartthings.apps()
        # assert
        assert len(apps) == 1

    @staticmethod
    def test_app(requests_mock):
        """Tests the app(id) method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        app = smartthings.app(api_mock.APP_ID)
        # Assert
        assert app.app_id == api_mock.APP_ID

    @staticmethod
    def test_create_app(requests_mock):
        """Tests the create app method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        app = App()
        data = get_json('app_post_request.json')
        data['appId'] = api_mock.APP_ID
        app.apply_data(data)
        # Act
        app, oauth = smartthings.create_app(app)
        # Assert
        assert app.app_id == api_mock.APP_ID
        assert oauth.client_id == '7cd4d474-7b36-4e03-bbdb-4cd4ae45a2be'
        assert oauth.client_secret == '9b3fd445-42d6-441b-b386-99ea51e13cb0'

    @staticmethod
    def test_delete_app(requests_mock):
        """Tests the delete app method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act/Assert
        result = smartthings.delete_app(api_mock.APP_ID)
        # Assert
        assert result

    @staticmethod
    def test_app_settings(requests_mock):
        """Tests retrieval of app settings."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        app_id = api_mock.APP_ID
        # Act
        settings = smartthings.app_settings(app_id)
        # Assert
        assert settings.app_id == app_id
        assert settings.settings == {'test': 'test'}

    @staticmethod
    def test_update_app_settings(requests_mock):
        """Tests updating app settings."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        app_id = api_mock.APP_ID
        settings = AppSettings(app_id)
        settings.settings['test'] = 'test'
        # Act
        entity = smartthings.update_app_settings(settings)
        # Assert
        assert entity.app_id == settings.app_id
        assert entity.settings == {'test': 'test'}

    @staticmethod
    def test_app_oauth(requests_mock):
        """Tests retrieval of OAuth settings."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        app_id = api_mock.APP_ID
        # Act
        oauth = smartthings.app_oauth(app_id)
        # Assert
        assert oauth.app_id == app_id
        assert oauth.client_name == 'pysmartthings-test'
        assert oauth.scope == ['r:devices']

    @staticmethod
    def test_update_app_oauth(requests_mock):
        """Tests updating OAuth settings."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        app_id = api_mock.APP_ID
        oauth = OAuth(app_id)
        oauth.client_name = 'pysmartthings-test'
        oauth.scope.append('r:devices')
        # Act
        oauth_entity = smartthings.update_app_oauth(oauth)
        # Assert
        assert oauth_entity.app_id == oauth.app_id
        assert oauth_entity.client_name == oauth.client_name
        assert oauth_entity.scope == oauth.scope

    @staticmethod
    def test_installedapps(requests_mock):
        """Tests the installedapps method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        apps = smartthings.installedapps()
        # Assert
        assert len(apps) == 2

    @staticmethod
    def test_installedapp(requests_mock):
        """Tests the installedapp(id) method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        app = smartthings.installedapp(api_mock.INSTALLED_APP_ID)
        # Assert
        assert app.installed_app_id == api_mock.INSTALLED_APP_ID

    @staticmethod
    def test_delete_installedapp(requests_mock):
        """Tests the delete app method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act/Assert
        result = smartthings.delete_installedapp(api_mock.INSTALLED_APP_ID)
        # Assert
        assert result

    @staticmethod
    def test_subscriptions(requests_mock):
        """Tests the get subscriptions method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        subscriptions = smartthings.subscriptions(api_mock.INSTALLED_APP_ID)
        # Assert
        assert len(subscriptions) == 3

    @staticmethod
    def test_delete_subscriptions(requests_mock):
        """Tests the delete subscriptions method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        count = smartthings.delete_subscriptions(api_mock.INSTALLED_APP_ID)
        # Assert
        assert count == 3

    @staticmethod
    def test_delete_subscription(requests_mock):
        """Tests the delete subscription method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        # Act
        deleted = smartthings.delete_subscription(
            api_mock.INSTALLED_APP_ID, api_mock.SUBSCRIPTION_ID)
        # Assert
        assert deleted

    @staticmethod
    def test_create_subscription(requests_mock):
        """Tests the create subscription method."""
        # Arrange
        api_mock.setup(requests_mock)
        smartthings = SmartThings(api_mock.API_TOKEN)
        sub = Subscription()
        sub.source_type = 'CAPABILITY'
        sub.location_id = api_mock.LOCATION_ID
        sub.capability = 'switch'
        sub.installed_app_id = api_mock.INSTALLED_APP_ID
        # Act
        entity = smartthings.create_subscription(sub)
        # Assert
        assert entity.subscription_id == api_mock.SUBSCRIPTION_ID



"""Define errors that can be returned from the SmartThings API."""

UNAUTHORIZED_ERROR = \
    "Authorization for the API is required, but the request has not been " \
    "authenticated."
FORBIDDEN_ERROR = \
    "The request has been authenticated but does not have appropriate" \
    "permissions, or a requested resource is not found."
UNKNOWN_ERROR = "An unknown API error occurred."


class APIUnauthorizedError(Exception):
    """Define the error when the API key is invalid."""

    def __init__(self):
        """Initialize instance of APIUnauthorizedError."""
        Exception.__init__(self, UNAUTHORIZED_ERROR)


class APIForbiddenError(Exception):
    """Define the error when the API key does not have permission."""

    def __init__(self):
        """Initialize instance of APIForbiddenError."""
        Exception.__init__(self, FORBIDDEN_ERROR)


class APIUnknownError(Exception):
    """Define a general error condition."""

    def __init__(self):
        """Initialize instance of APIUnknownError."""
        Exception.__init__(self, UNKNOWN_ERROR)


class APIInvalidGrant(Exception):
    """Define an invalid grant error."""

    pass



"""Tests for the Device file."""

import pytest

from pysmartthings.api import API
from pysmartthings.device import (
    Attribute, Device, DeviceEntity, DeviceStatus, DeviceType)

from . import api_mock
from .utilities import get_json


class TestDevice:
    """Tests for the Device class."""

    @staticmethod
    def test_init():
        """Tests whether the Device class initializes correctly."""
        # Arrange/Act
        device = Device()
        # Assert
        assert device.type == DeviceType.UNKNOWN
        assert device.capabilities == []
        assert device.components == {}

    @staticmethod
    def test_apply_data():
        """Tests the apply data method."""
        # Arrange
        data = get_json('device.json')
        device = Device()
        # Act
        device.apply_data(data)
        # Assert
        assert device.device_id == api_mock.DEVICE_ID
        assert device.name == 'GE In-Wall Smart Dimmer'
        assert device.label == 'Front Porch Lights'
        assert device.location_id == api_mock.LOCATION_ID
        assert device.type is DeviceType.DTH
        assert device.device_type_id == '8a9d4b1e3b9b1fe3013b9b206a7f000d'
        assert device.device_type_name == 'Dimmer Switch'
        assert device.device_type_network == 'ZWAVE'
        assert device.capabilities == [
            'switch', 'switchLevel', 'refresh', 'indicator', 'sensor',
            'actuator', 'healthCheck', 'light']
        assert device.components == {
            "main": [
                'switch', 'switchLevel', 'refresh', 'indicator', 'sensor',
                'actuator', 'healthCheck', 'light']
        }


class TestDeviceEntity:
    """Tests for the DeviceEntity class."""

    @staticmethod
    def test_refresh(requests_mock):
        """Tests the refresh method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        # Act
        device.refresh()
        # Assert
        assert device.label == 'Front Porch Lights'

    @staticmethod
    def test_save():
        """Tests the save method."""
        # Arrange
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            device.save()

    @staticmethod
    def test_switch_on(requests_mock):
        """Tests the switch_on method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        # Act
        result = device.switch_on()
        # Assert
        assert result
        assert not device.status.switch

    @staticmethod
    def test_switch_on_update(requests_mock):
        """Tests the switch_on method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        # Act
        result = device.switch_on(True)
        # Assert
        assert result
        assert device.status.switch

    @staticmethod
    def test_switch_off(requests_mock):
        """Tests the switch_on method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        device.status.switch = True
        # Act
        result = device.switch_off()
        # Assert
        assert result
        assert device.status.switch

    @staticmethod
    def test_switch_off_update(requests_mock):
        """Tests the switch_on method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        device.status.switch = True
        # Act
        result = device.switch_off(True)
        # Assert
        assert result
        assert not device.status.switch

    @staticmethod
    def test_set_level(requests_mock):
        """Tests the set_level method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        # Act
        result = device.set_level(75, 2)
        # Assert
        assert result
        assert device.status.level == 0

    @staticmethod
    def test_set_level_update(requests_mock):
        """Tests the set_level method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        device = DeviceEntity(api, device_id=api_mock.DEVICE_ID)
        # Act
        result = device.set_level(75, 2, True)
        # Assert
        assert result
        assert device.status.level == 75

    @staticmethod
    def test_status():
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(None, device_id=api_mock.DEVICE_ID)
        # Act
        status = device.status
        # Assert
        assert status.device_id == api_mock.DEVICE_ID


class TestDeviceStatus:
    """Tests for the DeviceStatus class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange/Act
        status = DeviceStatus(None, device_id=api_mock.DEVICE_ID)
        # Assert
        assert status.device_id == api_mock.DEVICE_ID
        assert status.attributes == {}
        assert not status.switch
        assert not status.motion
        assert status.level == 0

    @staticmethod
    def test_apply_data():
        """Tests the apply_data method."""
        # Arrange
        data = get_json('device_status.json')
        # Act
        status = DeviceStatus(None, api_mock.DEVICE_ID, data)
        # Assert
        assert len(status.attributes) == 9
        assert status.switch
        assert status.level == 100

    @staticmethod
    def test_apply_attribute_update():
        """Tests the apply_attribute_update method."""
        # Arrange
        data = get_json('device_status.json')
        status = DeviceStatus(None, api_mock.DEVICE_ID, data)
        # Act
        status.apply_attribute_update('main', 'switchLevel', 'level', 50)
        # Assert
        assert status.level == 50

    @staticmethod
    def test_refresh(requests_mock):
        """Tests the refresh method."""
        # Arrange
        api_mock.setup(requests_mock)
        api = API(api_mock.API_TOKEN)
        status = DeviceStatus(api, device_id=api_mock.DEVICE_ID)
        # Act
        status.refresh()
        # Assert
        assert len(status.attributes) == 9

    @staticmethod
    def test_switch():
        """Tests the init method."""
        # Arrange
        status = DeviceStatus(None, device_id=api_mock.DEVICE_ID)
        # Act
        status.switch = True
        # Assert
        assert status.switch

    @staticmethod
    def test_level():
        """Tests the init method."""
        # Arrange
        status = DeviceStatus(None, device_id=api_mock.DEVICE_ID)
        # Act
        status.level = 50
        # Assert
        assert status.level == 50

    @staticmethod
    def test_is_on():
        """Tests the is_on method."""
        # Arrange
        status = DeviceStatus(None, device_id=api_mock.DEVICE_ID)
        status.attributes[Attribute.acceleration] = 'active'
        status.attributes[Attribute.level] = 100
        # Act/Assert
        assert status.is_on(Attribute.acceleration)
        assert status.is_on(Attribute.level)
        assert not status.is_on(Attribute.switch)



"""SmartThings Cloud API"""

from setuptools import find_packages, setup

import pysmartthings

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name=pysmartthings.__title__,
      version=pysmartthings.__version__,
      description='A python library for interacting with the SmartThings cloud API.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/andrewsayre/pysmartthings',
      author='Andrew Sayre',
      author_email='andrew@sayre.net',
      license='MIT',
      packages=find_packages(),
      install_requires=['requests'],
      tests_require=['tox'],
      platforms=['any'],
      keywords="smartthings",
      zip_safe=False,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Topic :: Software Development :: Libraries",
          "Topic :: Home Automation",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          ])



"""Define the oauth module."""

from typing import List, Optional

from .api import API
from .entity import Entity
from .oauthapi import OAuthAPI


class OAuth:
    """Define the app OAuth settings."""

    _app_id: str
    _client_name: str
    _scope: List[str]

    def __init__(self, app_id: str):
        """Initialize a new instance of the OAuth class."""
        self._app_id = app_id
        self._client_name = None
        self._scope = []

    def to_data(self) -> dict:
        """Get a data representation of the instance."""
        return {
            'clientName': self._client_name,
            'scope': self._scope
        }

    def apply_data(self, data: dict):
        """Load the data of the instance."""
        self._client_name = data['clientName']
        self._scope = data.get('scope', self._scope)

    @property
    def app_id(self) -> str:
        """Get the app id the settings are associated with."""
        return self._app_id

    @property
    def client_name(self) -> str:
        """Get the name given to the OAuth Client."""
        return self._client_name

    @client_name.setter
    def client_name(self, value: str):
        """Set the name given to the OAuth client."""
        if not value:
            raise ValueError("Value can not be None or an empty string.")
        self._client_name = value

    @property
    def scope(self):
        """Get the ;ist of SmartThings API OAuth scope identifiers."""
        return self._scope


class OAuthEntity(Entity, OAuth):
    """Define oauth client settings."""

    def __init__(self, api: API, app_id: str, data=None):
        """Create a new instance of the OAuth class."""
        Entity.__init__(self, api)
        OAuth.__init__(self, app_id)
        if data:
            self.apply_data(data)

    def refresh(self):
        """Retrieve the latest values from the API."""
        data = self._api.get_app_oauth(self._app_id)
        if data:
            self.apply_data(data)

    def save(self):
        """Save changes to the app OAuth Client settings."""
        response = self._api.update_app_oauth(self._app_id, self.to_data())
        if response:
            self.apply_data(response)


class OAuthClient:
    """Define an oauth client information."""

    def __init__(self, data: Optional[dict]):
        """Create a new instance of the OAuthClient."""
        self._client_id = None
        self._client_secret = None
        if data:
            self.apply_data(data)

    def apply_data(self, data: dict):
        """Apply the given data to the entity."""
        self._client_id = data['oauthClientId']
        self._client_secret = data['oauthClientSecret']

    @property
    def client_id(self):
        """Get the client id."""
        return self._client_id

    @property
    def client_secret(self):
        """Get the client secret."""
        return self._client_secret


class OAuthToken:
    """Define oauth token information."""

    _api: OAuthAPI
    _access_token: str
    _refresh_token: str
    _expires_in: int
    _token_type: str
    _scope: List[str]

    def __init__(self, api: OAuthAPI, data: Optional[dict] = None,
                 refresh_token: Optional[str] = None):
        """Create a new instance of the OAuthToken class."""
        self._api = api
        self._access_token = None
        self._refresh_token = refresh_token
        self._expires_in = 0
        self._token_type = None
        self._scope = []
        if data:
            self.apply_data(data)

    def apply_data(self, data: dict):
        """Apply the given data to the entity."""
        self._access_token = data['access_token']
        self._token_type = data['token_type']
        self._refresh_token = data['refresh_token']
        self._expires_in = data['expires_in']
        data_scope = data['scope']
        if isinstance(data_scope, list):
            self._scope = data_scope
        if isinstance(data_scope, str):
            self._scope.clear()
            self._scope.append(data_scope)

    def refresh(self):
        """Refresh the auth and refresh tokens."""
        data = self._api.get_token(self._refresh_token)
        if data:
            self.apply_data(data)

    @property
    def access_token(self) -> str:
        """Get the access token for authentication."""
        return self._access_token

    @property
    def refresh_token(self) -> str:
        """Get the refresh token for obtaining new access tokens."""
        return self._refresh_token

    @property
    def expires_in(self) -> int:
        """Get the amount of time in seconds until the token expires."""
        return self._expires_in

    @property
    def token_type(self) -> str:
        """Get the type of token."""
        return self._token_type

    @property
    def scope(self) -> List[str]:
        """Get the scopes the token has permission to."""
        return self._scope



"""A python library for interacting with the SmartThings cloud API."""

from .smartthings import SmartThings

__title__ = "pysmartthings"
__version__ = "0.3.4"


def create(token):
    """Create an instance of the SmartThings API with the specified token."""
    return SmartThings(token)



"""Defines a SmartThings device."""

from enum import Enum
from typing import Any, Dict, Optional, Sequence

from .api import API
from .entity import Entity


class Attribute:
    """Define common attributes."""

    acceleration = 'acceleration'
    contact = 'contact'
    filter_status = 'filterStatus'
    level = 'level'
    motion = 'motion'
    mute = 'mute'
    presence = 'presence'
    sound = 'sound'
    switch = 'switch'
    tamper = 'tamper'
    valve = 'valve'
    water = 'water'


class Capability:
    """Define common capabilities."""

    light = 'light'
    motion_sensor = 'motionSensor'
    switch = 'switch'
    switch_level = 'switchLevel'


class Command:
    """Define common commands."""

    off = 'off'
    on = 'on'
    set_level = 'setLevel'


class DeviceType(Enum):
    """Define the device type."""

    UNKNOWN = 'UNKNOWN'
    DTH = 'DTH'
    ENDPOINT_APP = 'ENDPOINT_APP'


ATTRIBUTE_ON_VALUES = {
    Attribute.acceleration: 'active',
    Attribute.contact: 'open',
    Attribute.filter_status: 'replace',
    Attribute.motion: 'active',
    Attribute.mute: 'muted',
    Attribute.presence: 'present',
    Attribute.sound: 'detected',
    Attribute.switch: 'on',
    Attribute.tamper: 'detected',
    Attribute.valve: 'open',
    Attribute.water: 'wet'
}


class Device:
    """Represents a SmartThings device."""

    def __init__(self):
        """Initialize a new device."""
        self._device_id = None
        self._name = None
        self._label = None
        self._location_id = None
        self._type = DeviceType.UNKNOWN
        self._device_type_id = None
        self._device_type_name = None
        self._device_type_network = None
        self._components = {}
        self._capabilities = []

    def apply_data(self, data: dict):
        """Apply the given data dictionary."""
        self._device_id = data['deviceId']
        self._name = data['name']
        self._label = data['label']
        self._location_id = data['locationId']
        self._type = DeviceType(data['type'])
        self._components.clear()
        self._capabilities.clear()
        for comp_data in data['components']:
            capabilities = []
            for capability_data in comp_data['capabilities']:
                capability_id = capability_data['id']
                capabilities.append(capability_id)
                if id not in self._capabilities:
                    self._capabilities.append(capability_id)
            self._components[comp_data['id']] = capabilities
        if self._type is DeviceType.DTH:
            dth = data['dth']
            self._device_type_id = dth["deviceTypeId"]
            self._device_type_name = dth["deviceTypeName"]
            self._device_type_network = dth["deviceNetworkType"]

    @property
    def device_id(self) -> str:
        """Get the SmartThings device id."""
        return self._device_id

    @property
    def name(self) -> str:
        """Get the SmartThings device name."""
        return self._name

    @property
    def label(self) -> str:
        """Get the SmartThings user assigned label."""
        return self._label

    @property
    def location_id(self) -> str:
        """Get the SmartThings location assigned to the device."""
        return self._location_id

    @property
    def type(self) -> DeviceType:
        """Get the SmartThings device type."""
        return self._type

    @property
    def device_type_id(self) -> str:
        """Get the SmartThings device type handler id."""
        return self._device_type_id

    @property
    def device_type_name(self) -> str:
        """Get the SmartThings device type handler name."""
        return self._device_type_name

    @property
    def device_type_network(self) -> str:
        """Get the SmartThings device type handler network."""
        return self._device_type_network

    @property
    def components(self) -> Dict[str, Sequence[str]]:
        """Get the components of the device."""
        return self._components

    @property
    def capabilities(self) -> Sequence[str]:
        """Get a unique list of all capabilities across components."""
        return self._capabilities


class DeviceStatus:
    """Define the device status."""

    def __init__(self, api: API, device_id: str, data=None):
        """Create a new instance of the DeviceStatusEntity class."""
        self._api = api
        self._attributes = {}
        self._device_id = device_id
        if data:
            self.apply_data(data)

    @property
    def device_id(self) -> str:
        """Get the device id."""
        return self._device_id

    @device_id.setter
    def device_id(self, value: str):
        """Set the device id."""
        self._device_id = value

    def apply_data(self, data: dict):
        """Apply the values from the given data structure."""
        self._attributes.clear()
        for capabilities in data['components'].values():
            for attributes in capabilities.values():
                for attribute, value in attributes.items():
                    self._attributes[attribute] = value['value']

    def apply_attribute_update(self, component_id: str, capability: str,
                               attribute: str, value: Any):
        """Apply an update to a specific attribute."""
        # component_id and capability future usage.
        self._attributes[attribute] = value

    def is_on(self, attribute: str) -> bool:
        """Determine if a specific attribute contains an on/True value."""
        if attribute not in ATTRIBUTE_ON_VALUES:
            return bool(self._attributes.get(attribute))
        return self._attributes.get(attribute) == \
            ATTRIBUTE_ON_VALUES[attribute]

    def refresh(self):
        """Refresh the values of the entity."""
        data = self._api.get_device_status(self.device_id)
        if data:
            self.apply_data(data)

    @property
    def attributes(self):
        """Get all of the attribute values."""
        return self._attributes

    @property
    def switch(self) -> bool:
        """Get the switch attribute."""
        return self.is_on(Attribute.switch)

    @switch.setter
    def switch(self, value: bool):
        """Set the value of the switch attribute."""
        self._attributes[Attribute.switch] = \
            ATTRIBUTE_ON_VALUES[Attribute.switch] if value else 'off'

    @property
    def level(self) -> int:
        """Get the level attribute."""
        return int(self._attributes.get(Attribute.level, 0))

    @level.setter
    def level(self, value: int):
        """Set the level of the attribute."""
        self._attributes[Attribute.level] = value

    @property
    def motion(self) -> bool:
        """Get the motion attribute."""
        return self.is_on(Attribute.motion)


class DeviceEntity(Entity, Device):
    """Define a device entity."""

    def __init__(self, api: API, data: Optional[dict] = None,
                 device_id: Optional[str] = None):
        """Create a new instance of the DeviceEntity class."""
        Entity.__init__(self, api)
        Device.__init__(self)
        if data:
            self.apply_data(data)
        if device_id:
            self._device_id = device_id
        self._status = DeviceStatus(api, self._device_id)

    def refresh(self):
        """Refresh the device information using the API."""
        data = self._api.get_device(self._device_id)
        if data:
            self.apply_data(data)
        self._status.device_id = self._device_id

    def save(self):
        """Save the changes made to the device."""
        raise NotImplementedError

    def command(self, capability, command, args=None):
        """Execute a command on the device."""
        response = self._api.post_command(
            self._device_id, capability, command, args)
        return response == {}

    def switch_on(self, set_status: bool = False) -> bool:
        """Turn on the device."""
        result = self.command(Capability.switch, Command.on)
        if result and set_status:
            self.status.switch = True
        return result

    def switch_off(self, set_status: bool = False) -> bool:
        """Turn on the device."""
        result = self.command(Capability.switch, Command.off)
        if result and set_status:
            self.status.switch = False
        return result

    def set_level(self, level: int, duration: int,
                  set_status: bool = False) -> bool:
        """Set the level of the device."""
        result = self.command(
            Capability.switch_level, Command.set_level, [level, duration])
        if result and set_status:
            self.status.level = level
        return result

    @property
    def status(self):
        """Get the status entity of the device."""
        return self._status
